@@ -55,9 +55,6 @@ class PatchObject(object):
         self.patch_subject      = header['subject']
         self.patch_revision     = header['revision'] if header.has_key('revision') else 1
 
-        # self.extracted_patch    = None
-        self.unique_hash        = None
-
         self.filename           = filename
         self.offset_begin       = None
         self.offset_end         = None
@@ -305,6 +302,8 @@ def read_patch(filename):
         if r is not None: return "%s: %s" % (r.group(1), r.group(4)), int(r.group(3))
         r = re.match("^(.*) +%s$" % version, subject, re.IGNORECASE)
         if r is not None: return r.group(1).strip(), int(r.group(3))
+        r = re.match("^(.*)\\(resend\\)$", subject, re.IGNORECASE)
+        if r is not None: return r.group(1).strip(), 1
         return subject, 1
 
     header = {}
